/*! \file arhiplex_client.h
\brief include базовых arhiplex заголовочных файлов
    
Группировка include базовых header файлов в одном месте
*/

#include "arhiplex_cpp.h"
#include "arhiplex_itf.h"
#include "wrappers/operators.h"
#include "wrappers/utils.h"
#include "wrappers/model.h"
#include "wrappers/solve_result.h"
#include "wrappers/general_constraint.h"