/**
\private
*/
#pragma once
#include "../arhiplex_itf.h"
namespace arhiplex
{
inline
void releasable_object_deleter(IReleasable *obj)
{
    if(obj)
        obj->Release();
}

}; // namespace arhiplex