import arhiplexpy


if __name__ == "__main__":
    model = arhiplexpy.Model()

    model.read("blp-ic98.mps")
    model.set_log_file("arhiplex.log")

    result = model.solve()

    result.write_solution("blp-ic98.sol")