
#ifndef ARHIPLEX_CPP_EXPORT_H
#define ARHIPLEX_CPP_EXPORT_H

#ifdef ARHIPLEX_CPP_STATIC_DEFINE
#  define ARHIPLEX_CPP_EXPORT
#  define ARHIPLEX_CPP_NO_EXPORT
#else
#  ifndef ARHIPLEX_CPP_EXPORT
#    ifdef arhiplex_cpp_EXPORTS
        /* We are building this library */
#      define ARHIPLEX_CPP_EXPORT __attribute__((visibility("default")))
#    else
        /* We are using this library */
#      define ARHIPLEX_CPP_EXPORT __attribute__((visibility("default")))
#    endif
#  endif

#  ifndef ARHIPLEX_CPP_NO_EXPORT
#    define ARHIPLEX_CPP_NO_EXPORT __attribute__((visibility("hidden")))
#  endif
#endif

#ifndef ARHIPLEX_CPP_DEPRECATED
#  define ARHIPLEX_CPP_DEPRECATED __attribute__ ((__deprecated__))
#endif

#ifndef ARHIPLEX_CPP_DEPRECATED_EXPORT
#  define ARHIPLEX_CPP_DEPRECATED_EXPORT ARHIPLEX_CPP_EXPORT ARHIPLEX_CPP_DEPRECATED
#endif

#ifndef ARHIPLEX_CPP_DEPRECATED_NO_EXPORT
#  define ARHIPLEX_CPP_DEPRECATED_NO_EXPORT ARHIPLEX_CPP_NO_EXPORT ARHIPLEX_CPP_DEPRECATED
#endif

#if 0 /* DEFINE_NO_DEPRECATED */
#  ifndef ARHIPLEX_CPP_NO_DEPRECATED
#    define ARHIPLEX_CPP_NO_DEPRECATED
#  endif
#endif

#endif /* ARHIPLEX_CPP_EXPORT_H */
